#!/bin/env python
"""Generate sut java implementation from uppaal model"""
# Author: Harco Kuppens


import sys, os,  pprint, re, getpass, tempfile   # standard libraries
import argparse # only in standard lib for python 2.7 and later -> installed local copy  (pure python versions)
import jinja2, yaml  # external libraries   -> installed local copy   (pure python versions) (not yaml doesn't use libyaml but pure python version)
import pyuppaal  # local libraries          -> installed local copy   (pure python versions)
#import pygraphviz

import parser_types 
import model_parser




#-------------------------------------------------------------------------------
#   output
#-------------------------------------------------------------------------------

def write_sutinfo(output_file,data):
    import yaml  # external libraries   -> installed local copy   (pure python versions) (not yaml doesn't use libyaml but pure python version)

    selected_keys=('name','inputInterfaces','outputInterfaces','constants')
    #dictionary comprehension only supported for python 2.7 and higher
    #config={k:v for k,v in data.iteritems() if k in selected_keys}
    config={}
    for k,v in data.iteritems():
        if k in selected_keys:
            config[k]=v

    with open(output_file,'w') as f:
        f.write(yaml.safe_dump(config,default_flow_style=False))



def fill_template(output_file,data,template_dir,template_file):
    env = jinja2.Environment()
    env.loader = jinja2.FileSystemLoader(template_dir)
    tmpl = env.get_template(template_file)

    with open(output_file,'w') as f:
        f.write(tmpl.render(data))



#-------------------------------------------------------------------------------
#   main input parameters parsing
#-------------------------------------------------------------------------------


def isreadablefile(path):
    import os
    return  os.path.isfile(path) and os.access(path,os.R_OK)

def iswritablefile(path):
    import os
    return  os.path.isfile(path) and os.access(path,os.W_OK)


def iswritabledir(path):
    import os
    return  os.path.isdir(path) and os.access(path,os.W_OK)

def app_error(msg):
    print "\n\nAPPLICATION ERROR:\n"
    print msg
    exit(1)

#-------------------------------------------------------------------------------
#   main input parameters parsing
#-------------------------------------------------------------------------------


def get_models(models_dir):
    #TODO: find interactive :
    models=['ABP-output','BiometricPassport','Channel_Ack']
    return models


def get_models_help(models):
        help=[ "available models :" ]
        for model in models:
            help.append("   - " + model)
        return '\n'.join(help)

# python uppaalToJava\main.py modelFile classPathDir classPackage className

def args_parsing(models_dir):

    if models_dir:
       models=get_models(models_dir)

    #parser = argparse.ArgumentParser(prog=os.path.basename(sys.argv[0]),epilog=get_models_help(models))  # prints epilog always on single line
    parser = argparse.ArgumentParser(prog=os.path.basename(sys.argv[0]))

    #parser.add_argument('model', help='see below for available models')
    parser.add_argument('model', help='mode file in special uppaal xml format')
    parser.add_argument('classPathDir', help='directory which contains java classes in package subdirs')
    parser.add_argument('classPackage', help='package name of java class. eg. java.util')
    parser.add_argument('className', help='name of java class')
    parser.add_argument('sutInfoYamlFile', help='interfaces of sut in yaml file ')


    str=parser.format_help()

    def myhelp( ):
        #print str  + "\n" +  get_models_help(models)
        print str  + "\n"

    def myerror( message):
            print >>sys.stderr, ('%s: error: %s\n') % (parser.prog, message)
            parser.print_usage(sys.stderr)
            #print "\n" +  get_models_help(models)
            exit(1)

    if models_dir:
        parser.error=myerror
        parser.format_help=myhelp

    parse_result=parser.parse_args()


    model_file=parse_result.model

    if models_dir:
        # if only model name give (without .xml) then
        # we assume only modelname is given
        # so we create path from modelname using models_dir
        if not model_file.endswith('.xml'):
           model_name=model_file
           model_file= models_dir  + '/' + model_name + ".xml"

    model_file=os.path.abspath(model_file)

    # always use unix slashes in python!!
    # automatic convert any windows slashes
    # TODO: document this convention in python/txtdocs/
    model_file.replace('\\','/')





    classPathDir=os.path.abspath(parse_result.classPathDir)

    sutInfoYamlFile=os.path.abspath(parse_result.sutInfoYamlFile);


    # some checks on args
    #----------------------

    if not isreadablefile(model_file):
       app_error("cannot read model file : " + model_file )


    # classPathDir must already exist, but package subdirectories are created
    # on the fly
    if not iswritabledir(classPathDir) :
        app_error("cannot write to classPathDir : " +  classPathDir)

    return  (model_file,classPathDir,parse_result.classPackage,parse_result.className,sutInfoYamlFile)

def create_dir(dir_name):
    """ 
        create dir if not yes exists
    """
    import os
    if not os.path.isdir(dir_name):
        os.mkdir(dir_name)

def create_path(path_name):
    import os
    # determine subpaths by first finding absolute path
    # strip heading and tailing directory separators
    # split on directory separator
    subpaths=os.path.normpath(path_name).strip(os.sep).split(os.sep)

    if subpaths[0][1] == ":":
       # windows path starting with drive letter
       path= subpaths.pop(0) + os.sep
    else:
       # unix rootpath starting with separator
       path=os.sep

    while subpaths:
       path=path + os.sep + subpaths.pop(0)
       create_dir(path)


#-------------------------------------------------------------------------------
#   main
#-------------------------------------------------------------------------------

def javaFileString(classpath_dir,class_package,class_name):
    import os.path
    subpaths=class_package.split('.') 
    subpaths.append(class_name+".java")
    return os.path.join(classpath_dir,*subpaths)

def uppaal2sut(model_file,config_file,classpath_dir,class_package,class_name):
    parsed_data=model_parser.parse_uppaal_model(model_file)
    parsed_data['model_file']=model_file.replace('\\','\\\\')

    parsed_data['class_package']=class_package;
    parsed_data['class_name']=class_name;

    #pprint.pprint(parsed_data)
    if  config_file == "none":
        print "python: skipped writing config file"
    else:
        write_sutinfo(config_file,parsed_data)
        print "python: config written: " + config_file

    script_dir=os.path.abspath(os.path.dirname(sys.argv[0]))
    templates_dir=script_dir+'/templates/'

    sut_file=javaFileString(classpath_dir,class_package,class_name)


    fill_template(sut_file,parsed_data,templates_dir,'Sut.java')
    print "python: sut written: " + sut_file


def main():
    #note: dir already created by ruby or later by python java copy script!
    #sut_file =tmpdir + '/' + user + '/' + model_name + '/AbsLearnRGB/src/sut/Sut.java'



    # default_learning_params is also global variable


    #class_package="sut"
    #class_name="Sut"
    #classpath_dir=os.path.normpath("output/dynamic_classes")

    models_dir=""
    (model_file,classPathDir,classPackage,className,config_file)=args_parsing(models_dir)


    # create package subdirs
    subpaths=classPackage.split('.')
    class_dir= os.path.join(classPathDir,*subpaths)
    create_path(class_dir)

    #config_file=os.path.join(class_dir,'sutConfig.yaml')
    uppaal2sut(model_file,config_file,classPathDir,classPackage,className)


if __name__ == "__main__":
    main()



