"""
   parse uppaal model
"""
# Author: Harco Kuppens
import sys, os,  pprint, re, getpass, tempfile   # standard libraries
import argparse # only in standard lib for python 2.7 and later -> installed local copy  (pure python versions)
import jinja2, yaml  # external libraries   -> installed local copy   (pure python versions) (not yaml doesn't use libyaml but pure python version)
import pyuppaal  # local libraries          -> installed local copy   (pure python versions)
#import pygraphviz

import parser_types

import util

#-------------------------------------------------------------------------------
#    settings for parsing
#-------------------------------------------------------------------------------


# the default value an undefined state machine variable gets
statevar_undef_value=0


regexpr_simple_assignment=re.compile(r'([^\s]*)\s*=\s*(.*.)\s*')
regexpr_interface=re.compile(r'^void\s*(\w*)\s*\(([^)]*)\)')
regexpr_io_event=re.compile(r'^(\w*)\s*\(([^)]*)\)')



#-------------------------------------------------------------------------------
#   parsing  helpers
#-------------------------------------------------------------------------------


def strip_first_word_and_spaces(line):
    line=line.strip()
    return line[line.find(' ')+1:].strip()

def get_first_word(line):
    line=line.strip()
    return line[:line.find(' ')]

def isStringInteger(s):
    """
       returns true if string is an integer and else false
    """
    try:
        int(s)
        return True
    except ValueError:
        return False

def add_to_dict_with_list_value(dict,key,value):
    if not dict.has_key(key):
      dict[key]=[]
    dict[key].append(value)



#===============================================================================
#                                 parsing
#===============================================================================

#-------------------------------------------------------------------------------
#   model
#-------------------------------------------------------------------------------

def parse_uppaal_model(input_file):
    """
      parse uppaal model
    """

    # By default the sut name specified in sutinfo.yaml is derived
    # from the modelfilename. However if filename is model.xml
    # then the name of the parent folder is taken!
    name=util.get_bare_name(input_file)
    if name=="model":
        dirname=os.path.dirname(input_file)
        name=util.get_bare_name(dirname)
        

    nta=pyuppaal.from_xml(input_file)

    data=parse_declarations(nta.declaration)

    # add name of model
    data['name']=name

    template=nta.templates[0]
    data['locations']=[loc.id for loc in template.locations]
    data['initlocation']=template.initlocation.id

    # for input transitions
    #dictionary comprehension only supported for python 2.7 and higher
    #input_with_location2guard_with_action_with_target={ k:{} for k in data['inputInterfaces'].keys() }
    input_with_location2guard_with_action_with_target={}
    for k in data['inputInterfaces'].keys():
        input_with_location2guard_with_action_with_target[k]={}

    # for output transitions
    location2guard_with_action_with_output_with_target={}

    for transition in template.transitions:

       source=transition.source.id
       guard=transition.guard.get_value()
       action=transition.assignment.get_value()
       target=transition.target.id

       io_event=transition.synchronisation.get_value().strip()
       [io_event_name,io_type,param_names]=parse_io_event(io_event)


       # an action is ';' separated list of action statements
       action_statements=parse_action(action)

       # An input event can have each time differently labeled input
       # params which are used in the action statements of that transition.
       # For consistency we relabel all input variables to : p0,p1,p2...
       action_statements=[ relabel_params(stmt,param_names) for stmt in  action_statements]

       if guard:
           guard=relabel_params(guard,param_names)

       if io_type == 'input':

           trans_data={
             'guard':guard,
             'action_statements':action_statements,
             'target':target,
           }
           add_to_dict_with_list_value( input_with_location2guard_with_action_with_target[io_event_name] , source, trans_data )


       elif io_type == 'output':
           trans_data={
             'guard':guard,
             'action_statements':action_statements,
             'output_event': io_event_name,
             'output_event_params': param_names,
             'target':target,
           }
           add_to_dict_with_list_value(  location2guard_with_action_with_output_with_target , source , trans_data )

    data['input_with_location2guard_with_action_with_target']=input_with_location2guard_with_action_with_target

    data['location2guard_with_action_with_output_with_target']=location2guard_with_action_with_output_with_target
    return data

#-------------------------------------------------------------------------------
#   declarations
#-------------------------------------------------------------------------------

def parse_declarations(declarations_string):
    """
      parse declarations in uppaal model
    """
    data={'inputInterfaces':{},
          'outputInterfaces':{},
          'constants':[],
          'statevars':{}}


    # TODO: remove this old legacy structure:
    #       note: model parsing is now strict meaning that only a limited
    #             set of known param names may be used in declaration section!
    #             Below params are still in our models, so we must first clean them
    #             up before removing this piece of code
    data['configParams']={
      # learner
      'min_value':0,
      'max_value':255,
      'seed':1299777356020l,
      'maxNumTraces': 3000,
    	'minTraceLength':20,
    	'maxTraceLength':100,
      # lotos
      'lts_min_value':0,
      'lts_max_value':255,
      'lts_nat_type':'uint8_t',
      'lts_int_type':'int8_t'
    }

    #parse declarations
    for line in declarations_string.splitlines():
        if not line.strip():        # skip empty lines
           continue
        first_word=get_first_word(line).lower()
        first_var=first_word.split('=')[0].rstrip() # first_word: 'x=4' -> first_var:'x'
        if first_word == 'void':
            (interface,args)=parse_interface_declaration(line);
            if interface.startswith('I'):
                data['inputInterfaces'][interface] = args
            elif interface.startswith('O'):
                data['outputInterfaces'][interface] = args
            else:
                print "Interfaces must begin with 'I' for input or 'O' for output. : "  +  line
                exit(1)
        elif parser_types.is_type(first_word):
            javatype=parser_types.parse_type(first_word)
            decls=parse_statevar_declaration(line)
            for decl in decls:
              data['statevars'][ decl[0] ]= ( javatype, decl[1] )
        elif first_word == 'constants':
            data['constants']=parse_constants_declaration(line)
        elif first_var in data['configParams'].keys():
             pass
        elif first_word.startswith('//') or first_word.startswith('#'):
            pass
        elif line.strip() == '}':
            pass
        else:
           print "error cannot parse declaration line: " + line
           exit(1)

    return data




def parse_interface_declaration(statement):
    """
      parse interface declaration in uppaal model
    """
    inputInterfaces={}
    outputInterfaces={}
    match=regexpr_interface.match(statement)
    if match:
        interface = match.groups()[0]
        args=[]
        args_str = match.groups()[1]
        args_str.strip()
        if  args_str:
            for arg_str in args_str.split(','):
              type=arg_str.lstrip().split()[0]     # get first word
              parser_types.check_type(type)
              args.append(parser_types.parse_type(type))
    else:
        print "Error: failed parsing interface declaration at line: " +  statement
        exit(1)
    return [ interface, args ]


def parse_constants_declaration(statement):
    """
      parse constants declaration in uppaal model
    """
    statement=statement.rstrip(';') # strip  ; at end of statement
    value=parse_simple_assignment(statement)[1]
    constants=value.replace(' ','').split(',')
    return [int(x) for x in constants]                    # assume: all constants are integers!

def parse_statevar_declaration(statement):
    """
      parse a state variable declarations in uppaal model
    """
    decls=[]

    param_decls=strip_first_word_and_spaces(statement).rstrip(';').rstrip()
    param_decls=param_decls.split(',')
    for decl in param_decls:
        [name,value]=parse_simple_assignment(decl.strip())
        if value!='undef':
           value=int(value)                             # assume: all state values are integers!
        decls.append(  [name,value] )
    return decls


#-------------------------------------------------------------------------------
#  action
#-------------------------------------------------------------------------------



def parse_action(action):
    """
    parse an action, where an action is a ';' separated list of action statements
    """
    action_statements=action.strip().split(';')
    action_statements.pop()
    return action_statements



#-------------------------------------------------------------------------------
#  input/output event
#-------------------------------------------------------------------------------

def parse_io_event(io_event):
    """ 
       parse an input or output event on a transition in a uppaal model
    """  
    match=regexpr_io_event.match(io_event)
    if match:
        io_event_name = match.groups()[0].strip()
        args_str =match.groups()[1]
        param_names=[]
        if  args_str:
            param_names=args_str.strip().split(',')
            param_names=[param.strip() for param in param_names]
        if io_event_name.startswith('I'):
            io_type='input'
        elif io_event_name.startswith('O'):
            io_type='output'
        else:
            print "Error: failed parsing io event name; name must start with 'I' or 'O' " +  io_event_name
            exit(1)
    else:
        print "Error: failed parsing interface declaration at line: " +  statement
        exit(1)
    return [io_event_name,io_type,param_names]


def parse_simple_assignment(statement):
    """
      parse a simpel assignment
    """
    try:
        match=regexpr_simple_assignment.match(statement)
        if match:
           name=match.groups()[0]
           value=match.groups()[1]
        else:
           name=statement
           value=statevar_undef_value
    except Exception as inst:
        print "Error: failed parsing simple declaration at line: " +  statement
        exit(1)
    return [name,value]


#-------------------------------------------------------------------------------
#  relabel params in code 
#-------------------------------------------------------------------------------


def relabel_param(codestring,old_param,new_param):
    """
      relabel param in codestring  
    """
    import re
    codestring=re.sub(r'(\A|\W)' + old_param + r'(?=\W|\Z)',r'\1' + new_param,codestring)
    return codestring


def relabel_params(codestring,param_names):
    """
     An input event can have each time differently labeled input
     params which are used in the action statements of that transition.
     For consistency we relabel all input variables to : p0,p1,p2...
    """ 
    num_param=-1
    for param in param_names:
      num_param+=1
      if isStringInteger(param):
          continue
      codestring=relabel_param(codestring,param,'tomte_p'+str(num_param))
    return codestring





